/*
 *      $Id$
 *
 *      ProgJumper
 *
 * Created by René Hexel on 30/12/10.
 *
 * Copyright (c) 2010-2012 Rene Hexel. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by Rene Hexel.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * -----------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or
 * modify it under the above terms or under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/
 * or write to the Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */
#import <CoreMotion/CoreMotion.h>
#import "PROGGameViewController.h"
#import "Definitions.h"

#define PAN_FACTOR      128                     // pan dampening factor

#pragma mark - Private Category Interface

@interface PROGGameViewController ()

@property (assign) NSTimer *gameTimer;          /// timer that makes the game tick
@property (assign) PROGGameView *gameView;      /// main game view
@property (assign, getter=isGameRunning) BOOL gameRunning;   /// currently active?
@property (assign) BOOL switchingLevels;        /// am I between levels?
@property (assign) BOOL haveAccelerometer;      /// do I have an accelerometer?

- (void) gameOver;                              /// game is over

@end

#pragma mark - View Controller Implementation

@implementation PROGGameViewController
@synthesize delegate, gameScore, gameTimer, gameView, gameRunning, currentLevel;
@synthesize switchingLevels, motionManager, haveAccelerometer;

- (void) viewDidLoad
{
        [super viewDidLoad];

        gameView = (PROGGameView *) self.view;
        gameView.delegate = self;

        motionManager = [[CMMotionManager alloc] init];

        if (!currentLevel)                      // start with the first level
                self.currentLevel = [PROGLevel initialLevel];

        self.currentLevel.delegate = self;
}


- (void) viewDidUnload
{
        self.motionManager = nil;               // release
        self.gameView = nil;                    // avoid an expired pointer
        [self stopGame];
        [super viewDidUnload];
}


- (void) viewDidAppear: (BOOL) animated
{
        [super viewDidAppear: animated];

        /*
         * This is what makes the game tick:
         * Create a timer and tell it to send me a 'tick' message
         * for every frame (i.e. once every FRAMEPERIOD)
         */
        gameTimer = [NSTimer scheduledTimerWithTimeInterval: FRAMEPERIOD
                                                     target: self
                                                   selector: @selector(gameTick)
                                                   userInfo: self
                                                    repeats: YES];
        /*
         * initialise the accelerometer using Core Motion
         * (only works on a real device)
         */
        motionManager.accelerometerUpdateInterval = FRAMEPERIOD;
        [motionManager startAccelerometerUpdates];

        [self startGame];               // let's go
}


- (void) viewWillDisappear: (BOOL) animated
{
        [self stopGame];                // halt the game
        [gameTimer invalidate];         // this also releases the timer
        gameTimer = nil;                // so set it to nil

        [super viewWillDisappear: animated];
}


- (BOOL) shouldAutorotateToInterfaceOrientation: (UIInterfaceOrientation) interfaceOrientation
{
        /* small device (e.g. iPhone, iPod)? */
        if ([[UIDevice currentDevice] userInterfaceIdiom] == UIUserInterfaceIdiomPhone)
        {
                return interfaceOrientation == UIInterfaceOrientationPortrait;
        }

        /* big screen device, e.g. iPad */
        return interfaceOrientation == UIInterfaceOrientationPortrait ||
               interfaceOrientation == UIInterfaceOrientationPortraitUpsideDown;
}


#pragma mark - Game View delegate

- (void) drawRect: (CGRect) rect
{
        [currentLevel drawObjectsInRect: rect
                     withCurrentContext: UIGraphicsGetCurrentContext()];
}


#pragma mark - Game Play

- (IBAction) startGame
{
        gameRunning = YES;
        [currentLevel activate];
}


- (IBAction) stopGame
{
        gameRunning = NO;
        [currentLevel deactivate];
}


- (void) gameTick
{
        if (!gameRunning) return;               // ignore if game stopped

        CMAccelerometerData *accel = motionManager.accelerometerData;
        float x = accel.acceleration.x;
        float y = accel.acceleration.y;

        if (x != 0 || y != 0) haveAccelerometer = YES;

        if (haveAccelerometer)
                [currentLevel setAccelerationX: x  y: y];

        [currentLevel moveSprites];             // move foreground objects
        [currentLevel movePlayers];             // move players

        if (switchingLevels)
        {
                if ([currentLevel endOfLevelAnimationDone])
                {
                        NSArray *oldPlayers = [currentLevel.players retain];
                        self.currentLevel = [currentLevel nextLevel];
                        [currentLevel addPlayers: oldPlayers];
                        [oldPlayers release];
                        switchingLevels = NO;
                }
        }
        else
        {
                if ([currentLevel isLifeLost])          // did we just lose a life?
                {
                        // Insert code to handle loss of life
                        // at the moment, just end the game, but this needs changing!
                        [self gameOver];
                }
                else if ([currentLevel isFinished])     // are we done with this level?
                {
                        [self switchToNextLevel];

                        // Insert code to display the new level number
                }
        }

        [self.view setNeedsDisplay];            // redraw screen
}


- (void) switchToNextLevel
{
        switchingLevels = YES;                  // switching to next level
        [currentLevel startEndOfLevelAnimation];// level switch animation
}


- (void) gameOver
{
        [self stopGame];
        [delegate gameOverWithScore: gameScore];
}


#pragma mark - Level Delegate

- (CGRect) boundsForLevel: (PROGLevel *) level
{
        return self.view.bounds;
}

#pragma mark - Gesture recogniser delegate

- (IBAction) handlePanGesture: (UIPanGestureRecognizer *) sender
{
        if (sender.state == UIGestureRecognizerStateCancelled ||
            sender.state == UIGestureRecognizerStateFailed)
                return;

        CGPoint pan = [sender translationInView: self.view];

        [currentLevel setSpeedX: pan.x/PAN_FACTOR  y: pan.y/PAN_FACTOR];
}


#pragma mark - Memory Management

- (void) didReceiveMemoryWarning
{
        [super didReceiveMemoryWarning];

        // Release any cached data, images, etc. that aren't in use.
}


- (void) dealloc
{
        [motionManager release];
        [gameTimer invalidate];

        [super dealloc];
}

@end
