/*      $Id$
 *
 * Created by René Hexel on 18/12/09.
 *
 * Copyright (c) 2009-2012 Rene Hexel. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by Rene Hexel.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#import <math.h>
#import "PROGLevel.h"
#import "PROGSprite.h"
#import "PROGSound.h"
#import "Definitions.h"

#define PLAYER_IMAGE            @"Jumper.png"

#define BOUNCE_SOUND_NAME       @"Boing"
#define LOST_SOUND_NAME         @"Lost"

#pragma mark - PROGLevel private properties and methods

@interface PROGLevel ()

@property (nonatomic, retain) PROGSound *bounceSound;           /// bounce sound
@property (nonatomic, retain) PROGSound *lostSound;             /// life lost sound
@property (nonatomic, retain) PROGSound *vibrationSound;        /// iPhone vibration

@property (nonatomic, retain) NSArray *imageNames;              /// valid images for this level

@property (nonatomic, assign) BOOL animatingForEndOfLevel;      /// switching levels?

- (PROGSprite *) randomSprite;
- (void) setInitialPositionForSprite: (PROGSprite *) sprite;


@end

#pragma mark - PROGLevel implementation

@implementation PROGLevel

@synthesize delegate;
@synthesize levelNumber, lifeLost, finished;
@synthesize players, sprites, backgroundSprites, bounceSound;
@synthesize lostSound, vibrationSound;
@synthesize maxNumberOfSprites, maximumVelocityForLevel;
@synthesize imageNames, endOfLevelAnimationDone, animatingForEndOfLevel;

#pragma mark - Level and memory management

+ (PROGLevel *) initialLevel
{
        return [[[[self alloc] init] autorelease] nextLevel];
}


- init
{
        return [self initWithLevelNumber: 0];
}


- initWithLevelNumber: (NSInteger) currentLevelNumber
{
        if (!(self = [super init])) return nil;

        self.levelNumber = currentLevelNumber;

        [self setupLevel];

        return self;
}


- (void) setupLevel
{
        srandom(levelNumber);

        finished = NO;
        lifeLost = NO;
        animatingForEndOfLevel = NO;
        endOfLevelAnimationDone = NO;

        self.sprites = [NSMutableArray arrayWithCapacity: 128];
        self.backgroundSprites = [NSMutableArray arrayWithCapacity: 32];
        self.players = [NSMutableArray arrayWithCapacity: 8];

        self.bounceSound = [PROGSound soundNamed: BOUNCE_SOUND_NAME];
        self.lostSound = [PROGSound soundNamed: LOST_SOUND_NAME];
        self.vibrationSound = [PROGSound vibrationSound];

        self.imageNames = [self imageNamesForCurrentLevel];
}


- (void) dealloc
{
        [vibrationSound release];
        [bounceSound release];
        [lostSound release];

        [backgroundSprites release];
        [sprites release];
        [players release];
        [imageNames release];

        [super dealloc];
}


- (PROGLevel *) nextLevel
{
        NSString *className = [NSString stringWithFormat: @"PROGLevel_%d",
                               ++levelNumber];
        Class levelClass = NSClassFromString(className);

        /*
         * try to allocate a new level with the class name based on
         * the level number.  If no new class exists, just stick with
         * the current instance (the level number has already been
         * incremented).
         */
        PROGLevel *nextLevel = [[levelClass alloc]
                                initWithLevelNumber: levelNumber];
        if (nextLevel)
        {
                nextLevel.delegate = self.delegate;

                return [nextLevel autorelease];
        }

        /*
         * no new level class? --> just stick with the current level instance
         * with the level number incremented above
         */
        [self setupLevel];

        return self; 
}


- (NSArray *) imageNamesForCurrentLevel
{
        return [NSArray array]; // subclass needs to implement this properly!
}


#pragma mark - Game play

- (void) activate   { /* must be handled in a concrete level subclass */ }

- (void) deactivate { /* must be handled in a concrete level subclass */ }

- (void) setInitialPositionForSprite: (PROGSprite *) sprite
{
        long mv = maximumVelocityForLevel;

        CGRect bounds = [delegate boundsForLevel: self];
        CGSize size = sprite.dimensions.size;

        CGPoint velocity =
        {
                mv ? (random() % mv / FPS) : 0,         // horizontal speed
                0                                       // vertical speed
        };
        CGPoint position =
        {
                bounds.origin.x + random() % (int) (bounds.size.width - size.width),
                bounds.origin.y + random() % (int) (bounds.size.height- size.height)
        };

        sprite.position = position;
        sprite.velocity = velocity;
}

- (void) setSpeedX: (float) x  y: (float) y
{
        for (PROGSprite *player in players)
        {
                CGPoint speed = player.velocity;
                speed.x = x;
                player.velocity = speed;
        }
}


- (void) setAccelerationX: (float) x  y: (float) y
{
        for (PROGSprite *player in players)
        {
                CGPoint acceleration = player.acceleration;
                acceleration.x = x;
                player.acceleration = acceleration;
        }
}


- (void) setAccelerationForPlayer: (PROGSprite *) player
{
        player.acceleration = CGPointMake(0, levelNumber);
}

- (void) setInitialPositionForPlayer: (PROGSprite *) player
{
        PROGSprite *lowestSprite = player;
        int n = [sprites count];
        if (n)
        {
                for (PROGSprite *sprite in sprites)
                        if (sprite.dimensions.origin.y > lowestSprite.dimensions.origin.y)
                                lowestSprite = sprite;
                player.velocity = lowestSprite.velocity;
        }
        else
        {
                [self setInitialPositionForSprite: player];
                player.velocity = CGPointMake(0, 0);
        }

        CGPoint position = lowestSprite.position;

        position.y -= player.dimensions.size.height;

        player.position = position;

        [self setAccelerationForPlayer: player];
}


- (PROGSprite *) randomSprite
{
        NSInteger n = [imageNames count];
        NSInteger i = random() % n;
        NSString *imageName = [imageNames objectAtIndex: i];

        BOOL collision = NO;
        PROGSprite *sprite = [[PROGSprite alloc] initWithImageNamed: imageName];
        do
        {
                [self setInitialPositionForSprite: sprite];
                collision = NO;
                for (PROGSprite *other in sprites)
                        if ([sprite hasCollisionWith: other])
                                collision = YES;
        }
        while (collision);

        return [sprite autorelease];
}


- (void) addSprite
{
        if ([sprites count] < self.maxNumberOfSprites)
                [sprites addObject: [self randomSprite]];
}


- (void) addPlayer
{
        PROGSprite *player = [[PROGSprite alloc] initWithImageNamed: PLAYER_IMAGE];

        [self setInitialPositionForPlayer: player];
        [players addObject: player];
        [player release];
}


- (void) addPlayers: (NSArray *) playerArray
{
        for (PROGSprite *player in playerArray)
        {
                [self performSelector: @selector(setAccelerationForPlayer:)
                           withObject: player
                           afterDelay: 1.5];
                [players addObject: player];
        }
}


- (void) moveSprites
{
        CGRect bounds = [delegate boundsForLevel: self];
        BOOL fin = animatingForEndOfLevel;
        for (PROGSprite *sprite in sprites)
        {
                [sprite animate];
                fin &= sprite.dimensions.origin.y > bounds.origin.y + bounds.size.height;
        }
        if (fin)
                endOfLevelAnimationDone = YES;
}


- (BOOL) movePlayers
{
        BOOL collision = NO;

        for (PROGSprite *player in players)
        {
                [player animate];
                /*
                 * only check for bounce when coming down
                 */
                if (player.velocity.y > 0)
                {
                        CGRect bounds = [delegate boundsForLevel: self];
                        float px = player.dimensions.origin.x + player.dimensions.size.width / 2;
                        float py = player.dimensions.origin.y + player.dimensions.size.height;

                        if (!animatingForEndOfLevel) for (PROGSprite *sprite in sprites)
                        {
                                float sx = sprite.dimensions.origin.x;
                                float sy = sprite.dimensions.origin.y;
                                float sw = sprite.dimensions.size.width;
                                float sh = sprite.dimensions.size.height;

                                if (px > sx && px < sx + sw &&
                                    py > sy && py < sy + sh)    // bounce?
                                {
                                        if (player.velocity.y > 1) [bounceSound play];
                                        [player handleCollisionWith: sprite];
                                        collision = YES;
                                }
                        }
                        else if (py > 3*bounds.size.height/4)
                        {
                                player.acceleration = CGPointZero;
                                player.velocity = CGPointZero;
                        }
                }
        }

        return collision;
}


- (void) animateSpriteForEndOfLevel: (PROGSprite *) sprite
{
        sprite.acceleration = CGPointMake(0, 2);
        sprite.velocity = CGPointMake(0, 2);
        sprite.angularVelocity = 0.02;
}


- (void) animatePlayerForEndOfLevel: (PROGSprite *) player
{
        player.acceleration = CGPointMake(0, 2);
        player.velocity = CGPointMake(0, 2);
}




- (void) startEndOfLevelAnimation
{
        for (PROGSprite *sprite in backgroundSprites)
                [self animateSpriteForEndOfLevel: sprite];

        for (PROGSprite *sprite in sprites)
                [self animateSpriteForEndOfLevel: sprite];
        
        for (PROGSprite *player in players)
                [self animatePlayerForEndOfLevel: player];

        animatingForEndOfLevel = YES;
}


- (void) drawObjectsInRect: (CGRect) rect
        withCurrentContext: (CGContextRef) context
{
        for (PROGSprite *sprite in backgroundSprites)
        {
                if ([sprite intersectsWithRect: rect])          // visible?
                        [sprite drawInContext: context];
                else if (!animatingForEndOfLevel)
                        [self setInitialPositionForSprite: sprite];
        }

        for (NSUInteger i = 0; i < [sprites count]; i++)
        {
                PROGSprite *sprite = [sprites objectAtIndex: i];

                if ([sprite intersectsWithRect: rect])          // visible?
                        [sprite drawInContext: context];
                else if (!animatingForEndOfLevel)
                        [self setInitialPositionForSprite: sprite];
        }

        for (PROGSprite *player in players)
        {
                if ([player intersectsWithRect: rect])
                        [player drawInContext: context];
                else if (player.dimensions.origin.y + player.dimensions.size.height > rect.origin.y)
                {
                        lifeLost = [players count] <= 1;
                        [players performSelectorOnMainThread: @selector(removeObject:)
                                                  withObject: player
                                               waitUntilDone: NO];
                        [lostSound play];
                }
                else finished = YES;
        }
}

@end
