/*      $Id$
 *
 * Created by René Hexel on 23/03/08.
 *
 * Copyright (c) 2008-2012 Rene Hexel. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by Rene Hexel.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#import <math.h>
#import "Definitions.h"
#import "PROGSprite.h"

#define TERMINAL_VELOCITY       20

#pragma mark - private properties and Methods

@interface PROGSprite ()

@property (nonatomic, retain) NSDate *creationDate;

@end

#pragma mark - sprite implementation

@implementation PROGSprite

@synthesize image, dimensions, velocity, angularVelocity, angle, lifeTime;
@synthesize acceleration, alpha, creationDate;

#pragma mark - Memory Management

- initWithImageNamed: (NSString *) file
{
        UIImage *img = [UIImage imageNamed: file];

        return [self initWithImage: img];
}


- initWithImage: (UIImage *) anImage
{
        if (!(self = [super init]))
                return nil;
        
        [image CGImage];        // cache image
        
        if ((image = [anImage retain]) != nil)
                dimensions.size = [image size];
        else                    // image missing?
        {
                [self release];
                return nil;
        }

        self.creationDate = [NSDate date];

        return self;
}


- (void) dealloc
{
        [image release];
        [creationDate release];

        [super dealloc];
}

#pragma mark - manually implemented properties

- (CGPoint) position
{
        return dimensions.origin;
}


- (void) setPosition: (CGPoint) position
{
        dimensions.origin = position;
}


#pragma mark - Drawing

- (void) draw
{
        [self drawInContext: UIGraphicsGetCurrentContext()];
}


- (void) drawInContext: (CGContextRef) ctx
{
        CGRect imageRect = dimensions;
        CGPoint offset = { dimensions.size.width / 2, dimensions.size.height / 2 };

        /*
         * always draw/rotate around the centre point
         */
        imageRect.origin.x = -offset.x;
        imageRect.origin.y = -offset.y;

        /*
         * move into position and rotate
         */
        CGContextSaveGState(ctx);
        CGContextTranslateCTM(ctx, dimensions.origin.x + offset.x, dimensions.origin.y + offset.y);
        CGContextRotateCTM(ctx, angle);

        [image drawInRect: imageRect];

        /*
         * restore drawing context to its original state
         */
        CGContextRestoreGState(ctx);
}


#pragma mark -
#pragma mark Animation and interaction with other Sprites

- (void) animate
{
        dimensions.origin.x += velocity.x;
        dimensions.origin.y += velocity.y;

        /*
         * rotate and wrap around 360 degrees (2 * pi)
         */
        angle = fmodf(angle + angularVelocity, 2 * M_PI);

        /*
         * accelerate
         */
        velocity.x += acceleration.x;
        velocity.y += acceleration.y;

        if (velocity.x > TERMINAL_VELOCITY) velocity.x = TERMINAL_VELOCITY;
        if (velocity.y > TERMINAL_VELOCITY) velocity.y = TERMINAL_VELOCITY;

        /*
         * fade out after life time has expired
         */
        NSDate *now = [NSDate date];
        NSTimeInterval interval = [now timeIntervalSinceDate: creationDate];
        if (lifeTime && interval > lifeTime)
        {
                if (alpha > 0) alpha -= 0.1;    // fade out
                else self.creationDate = now;   // reset life time
        }
}


- (BOOL) intersectsWithRect: (CGRect) rect
{
        return INTERSECTS(dimensions, rect);
}


- (BOOL) hasCollisionWith: (PROGSprite *) otherObject
{
        CGRect otherObjectDimensions = otherObject.dimensions;

        return INTERSECTS(dimensions, otherObjectDimensions);
}


- (BOOL) handleCollisionWith: (PROGSprite *) otherObject
{
        velocity.y = otherObject.velocity.y - velocity.y;       // bounce up

        return YES;
}



@end
